import logging
import Queue
import threading
import time

from flask import current_app as app

from lpreports import db
from lpreports.launchpad import client as lp_client

log = logging.getLogger(__name__)

NUM_THREADS = 2
PROJECTS_LIST = ["fuel", "mos"]


class ProducerThread(threading.Thread):
    def __init__(self, queue, stop_event, lp, project, modified_since):
        super(ProducerThread, self).__init__()
        self.queue = queue
        self.stop_event = stop_event
        self.lp = lp
        self.project = project
        self.modified_since = modified_since

    def run(self):
        try:
            for bug in self.lp.find_bug_tasks(self.project,
                                              self.modified_since):
                self.queue.put(bug)
        except Exception as e:
            log.error(e, exc_info=True)
        finally:
            log.info("Producer finished produsing "
                     "bug tasks for {0}".format(self.project))
            self.stop_event.set()


class ConsumerThread(threading.Thread):
    def __init__(self, queue, stop_event, lp, project, bugs_db):
        super(ConsumerThread, self).__init__()
        self.queue = queue
        self.stop_event = stop_event
        self.lp = lp
        self.project = project
        self.bugs_db = bugs_db

    def run(self):
        while True:
            try:
                bug = self.queue.get(block=False)
                try:
                    processed_bug = self.lp.process_bug_task(bug)

                    target_name = processed_bug["target_name"]
                    if target_name in PROJECTS_LIST:
                        log.info("Adding bug #{0} - {1} to [{2} : {3}]".format(
                            processed_bug["id"],
                            processed_bug["title"].encode("ascii", "ignore"),
                            processed_bug["target_name"],
                            processed_bug["milestone"]))

                        db.insert_bug(processed_bug, self.bugs_db)
                except Exception as e:
                    log.error(e, exc_info=True)
                    log.error(str(bug))
            except Queue.Empty:
                if self.stop_event.is_set():
                    log.info("Consumer finished processing "
                             "bugs for {0}".format(self.project))
                    break
                time.sleep(0.2)


def fetch_bugs(projects, threads_num, modified_since):
    lp = lp_client.LaunchpadClient(app.config["LP_CONSUMER_KEY"],
                                   app.config["LP_TOKEN"],
                                   app.config["LP_SIGNATURE"])

    producers = []
    consumers = []

    for project in projects:
        queue = Queue.Queue()
        stop_event = threading.Event()

        producer = ProducerThread(queue, stop_event, lp,
                                  project, modified_since)
        producers.append(producer)

        for i in xrange(threads_num + 1):
            consumer = ConsumerThread(queue, stop_event,
                                      lp, project,
                                      db.bugs_db())
            consumers.append(consumer)

    for producer in producers:
        producer.start()

    for consumer in consumers:
        consumer.start()

    for consumer in consumers:
        consumer.join()


def syncdb():
    sync_start_time = time.time()

    lp = lp_client.LaunchpadClient(app.config["LP_CONSUMER_KEY"],
                                   app.config["LP_TOKEN"],
                                   app.config["LP_SIGNATURE"])

    mos = lp.get_project("mos")
    fuel = lp.get_project("fuel")

    milestones_list = list(set(mos["all_milestones"]) &
                           set(fuel["all_milestones"]))

    active_milestones_list = list(set(mos["active_milestones"]) &
                                  set(fuel["active_milestones"]))

    active_milestones_target_dates = mos["active_milestones_target_dates"]
    active_milestones_target_dates.update(fuel["active_milestones_target_dates"])

    log.info("Creating milestones...")
    db.upsert_milestones(milestones_list,
                         active_milestones_list,
                         active_milestones_target_dates)

    fetch_bugs(PROJECTS_LIST, NUM_THREADS, db.get_update_time())

    db.set_update_time(sync_start_time)

    lp.close_session()
