import collections
import json
import logging

import cachetools
from flask import current_app as app
from flask.ext.pymongo import PyMongo

from lpreports.launchpad.bug import Bug
import lpreports.utils.common

log = logging.getLogger(__name__)

bugs = None
assignees = None
users = None
security = None
cve = None


def init_app(app):
    global bugs, assignees, users, security, cve
    bugs = PyMongo(app, config_prefix='MONGO_BUGS')
    assignees = PyMongo(app, config_prefix='MONGO_ASSIGNEES')
    users = PyMongo(app, config_prefix='MONGO_USERS')
    security = PyMongo(app, config_prefix='MONGO_SECURITY')
    cve = PyMongo(app, config_prefix='MONGO_CVE')


def bugs_db():
    return bugs.db


def assignees_db():
    return assignees.db


def users_db():
    return users.db


def security_team_db():
    return security.db


def cve_db():
    return cve.db


def get_update_timestamp():
    try:
        ts = int(bugs_db().update_date.find_one()["Update_date"])
    except Exception:
        ts = 0
        log.warn("Failed to get last update time. "
                 "Defaulting to {0}".format(ts))

    return ts


def get_update_time():
    ts = get_update_timestamp()
    return lpreports.utils.common.timestamp_to_utc_date(ts)


def set_update_time(update_time):
    bugs_db().drop_collection("update_date")
    bugs_db().create_collection("update_date")
    bugs_db().update_date.insert({"Update_date": update_time})


def find_user(**kwargs):
    store = users_db().users
    return store.find_one(**kwargs)


def add_user(user, update=False):
    store = users_db().users

    if update:
        store.update({"_id": user["_id"]}, user)
    else:
        store.insert(user)


def get_user_security_teams_membership(email):
    store = security_team_db().people
    user = store.find_one({"email": email})

    if user:
        return user["projects"]
    else:
        return []


def get_milestones(active_only=False):
    id = 2 if active_only else 1
    raw = bugs_db().milestones.find_one({"_id": id})["Milestone"]
    major = [m for m in raw if "-" not in m]
    updates = [m for m in raw if "-" in m]

    return sorted(major, reverse=True) + sorted(updates, reverse=True)


def get_maintenance_milestones():
    return [ms for ms in get_milestones(active_only=True)
            if ms not in app.config['KEY_MILESTONES']]


def get_milestones_target_dates():
    return bugs_db().milestones.find_one({"_id": 3})["Milestone"]


def upsert_milestones(milestones, active_milestones,
                      active_milestones_target_dates):
    bugs_db().milestones.remove({"_id": 1})
    bugs_db().milestones.insert(
        {"_id": 1, "Milestone": sorted(milestones, reverse=True)})
    bugs_db().milestones.remove({"_id": 2})
    bugs_db().milestones.insert(
        {"_id": 2, "Milestone": sorted(active_milestones, reverse=True)})
    bugs_db().milestones.remove({"_id": 3})
    bugs_db().milestones.insert(
        {"_id": 3, "Milestone": active_milestones_target_dates},
        check_keys=False)


def get_major_teams():
    teams = collections.OrderedDict()

    for mteam in assignees_db().major_teams.find():
        teams[mteam['Team']] = mteam['Members']

    teams['Unknown'] = {'unknown': []}

    return teams


def get_full_teams():
    raw_data = assignees_db().full_teams.find_one()['Full teams']
    teams_data = json.loads(raw_data,
                            object_pairs_hook=collections.OrderedDict)
    teams_data['Unknown'] = {'unknown': []}
    return teams_data


def get_teams():
    full_teams = get_full_teams()
    result = collections.OrderedDict()
    for department, teams in full_teams.items():
        result[department] = sorted(teams.keys())
    return result


def get_department_members(department):
    """Get all members of given Mirantis department."""
    return assignees_db().major_teams.find_one({"Team": department})['Members']


def insert_bug(bug, _bugs_db):
    target_name = bug["target_name"]
    _bugs_db[target_name].update(
        {"$and": [
            {"id": bug['id']},
            {"milestone": bug["milestone"]}]},
        bug,
        upsert=True)


def delete_bug_task(project, task):
    bugs_db()[project].remove(task)


@cachetools.ttl_cache()
def get_all_bugs():
    return get_all_bugs_in_project("mos") + get_all_bugs_in_project("fuel")


@cachetools.ttl_cache()
def get_all_bugs_in_project(project):
    return [bug for bug in bugs_db()[project].find()]


def get_bugs(project_name, statuses, milestone_names=[],
             tags=None, importance=None, **kwargs):
    if not tags:
        tags = []

    if not importance:
        importance = []

    project = bugs_db()[project_name]

    search = [{"status": {"$in": statuses}}]

    if milestone_names:
        search.append({'milestone': {"$in": milestone_names}})

    if importance:
        search.append({"importance": {"$in": importance}})

    if tags:
        if kwargs.get("condition"):
            search.append({"tags": {"$nin": tags}})
        else:
            search.append({"tags": {"$in": tags}})

    return [Bug(r) for r in project.find({"$and": search})]


def get_cves():
    return cve_db().cves.find()
