import collections
import logging
import random
import six
import time

import cachetools
import iso8601
import requests
from requests import adapters

from lpreports.launchpad import consts
from lpreports.utils.common import select_keys

log = logging.getLogger(__name__)

LP_API = "https://api.launchpad.net/1.0"
LP_REQUESTS_RETRIES = 3

BUG_STATUSES = consts.BUG_STATUSES["All"]
DATE_FIELDS = consts.DATE_FIELDS


class LaunchpadClient(object):
    def __init__(self,
                 oauth_consumer_key,
                 oauth_token,
                 oauth_signature):
        self.oauth_consumer_key = oauth_consumer_key
        self.oauth_token = oauth_token
        self.oauth_signature = oauth_signature

        self.session = requests.Session()
        self.session.mount(
            LP_API,
            adapters.HTTPAdapter(max_retries=LP_REQUESTS_RETRIES))

    def close_session(self):
        log.info("Closing LP session")
        self.session.close()

    def get_project(self, project_name):
        url = "{0}/{1}".format(LP_API, project_name)
        raw = self._get_json_from_url(url)

        project = select_keys(raw, ["display_name", "name"])
        project["active_milestones"] = self.get_active_milestones(project_name)
        project["all_milestones"] = self.get_all_milestones(project_name)
        project["active_milestones_target_dates"] = self.get_active_milestones_target_dates(
            project_name,
            project["active_milestones"])
        return project

    def get_active_milestones(self, project_name):
        return self._generic_get_milestones(project_name, "active")

    def get_all_milestones(self, project_name):
        return self._generic_get_milestones(project_name, "all")

    def _generic_get_milestones(self, project_name, collection_type):
        url = "{0}/{1}/{2}_milestones".format(LP_API,
                                              project_name,
                                              collection_type)
        raw = self._get_json_from_url(url)
        return [m["name"] for m in raw["entries"]]

    def get_active_milestones_target_dates(self, project_name, active_milestones):
        result = {}
        for active_milestone in active_milestones:
            url = "{0}/{1}/+milestone/{2}".format(
                LP_API, project_name, active_milestone)
            milestone_data = self._get_json_from_url(url)
            date_targeted = milestone_data['date_targeted']
            if date_targeted:
                result[active_milestone] = self._process_lp_date(date_targeted)
        return result

    def get_milestone2series_mapping(self, project_name):
        url = "{0}/{1}/all_milestones".format(LP_API,
                                              project_name)
        raw = self._get_json_from_url(url)
        milestone_series = {}
        for m in raw["entries"]:
            milestone_series[m["self_link"]] = m["series_target_link"]

        milestone_series[None] = None
        return milestone_series

    def get_team_members(self, team):
        url = "{0}/~{1}/members".format(LP_API, team)
        raw = self._get_json_from_url(url)
        members = []

        for entry in raw["entries"]:
            name = entry["name"]
            display_name = entry["display_name"]
            email_link = entry["preferred_email_address_link"]
            email = LaunchpadClient._get_last_chunk_of_url(email_link)

            members.append({"name": name,
                            "display_name": display_name,
                            "email": email})

        return members

    @staticmethod
    def _construct_search_url(project, modified_since):
        log.info("Starting to search bug tasks in {0} "
                 "modified since {1}".format(project, modified_since))

        url = "{0}/{1}/".format(LP_API, project)
        url += "?ws.op=searchTasks"
        for status in BUG_STATUSES:
            url += '&status=' + six.moves.urllib.parse.quote_plus(status)

        if modified_since:
            url += "&modified_since={0}".format(modified_since)

        return url

    def find_bug_tasks(self, project, last_update_ts):
        url = LaunchpadClient._construct_search_url(project, last_update_ts)

        milestone_series = self.get_milestone2series_mapping(project)

        while url:
            log.debug("Getting chunk {0}".format(url))
            chunk = self._get_json_from_url(url)

            if not chunk:
                log.warning("No data from chunk {0}".format(url))
                break

            for record in chunk["entries"]:
                related_tasks_url = record["related_tasks_collection_link"]
                related_tasks_raw = self._get_json_from_url(related_tasks_url)
                related_tasks = related_tasks_raw["entries"]

                bug_milestone = record.get("milestone_link", None)
                rt_milestones = [rt["target_link"]
                                 for rt in related_tasks]

                if related_tasks:
                    # NOTE: this is a workaround for stupid Launchpad API
                    # which returns non-existing bug tasks. It actually might
                    # return a task for specific milestone which doesn't exist
                    # Such bugs would have "Status tracked in X.0.x" on UI
                    if milestone_series.get(bug_milestone) not in rt_milestones:
                        yield record

                    for related_task in related_tasks:
                        yield related_task
                else:
                    yield record

            url = chunk.get('next_collection_link')

        log.info("Finished searching bugs in {0}".format(project))

    def get_bug_by_id(self, bug_id):
        url = "{0}/bugs/{1}".format(LP_API, bug_id)
        return self._get_json_from_url(url)

    def get_bug_by_url(self, bug_link):
        return self._get_json_from_url(bug_link)

    def get_bug_verifier(self, bug_target_name, activity_link):
        # look through the activities in reversed order, find the last
        # person who moved bug to 'Fix Released' and consider that person
        # to be bug verifier
        activity_log = self._get_json_from_url(activity_link)

        for entry in reversed(activity_log["entries"]):
            person_link = entry.get("person_link", None)
            new_value = entry["newvalue"]
            what_changed = entry["whatchanged"]
            changed_bug_target = what_changed.split(":")[0]

            if (new_value == "Fix Released"
                    and what_changed.endswith("status")
                    and bug_target_name == changed_bug_target):
                return person_link

        log.warning("No verifier found for bug {0}".format(activity_link))
        return None

    def _process_bug_verifier(self, task, bug):
        if task["status"] == "Fix Released":
            verifier = self.get_bug_verifier(
                task["bug_target_name"],
                bug["activity_collection_link"])

            return self._get_last_chunk_of_url(verifier, delimiter="~")
        else:
            return None

    def _process_bug_dates(self, task):
        for field in DATE_FIELDS:
            val = task.get(field, None)
            if val:
                task[field] = self._process_lp_date(val)

    def _get_bug_cves(self, bug_id):
        url = "{0}/bugs/{1}/cves".format(LP_API, bug_id)
        raw_data = self._get_json_from_url(url)
        entries = raw_data["entries"]

        return [entry["display_name"] for entry in entries]

    def process_bug_task(self, raw):
        keys = ["status",
                "importance"]
        task = select_keys(raw, DATE_FIELDS + keys)

        task["assignee"] = self._get_last_chunk_of_url(raw["assignee_link"],
                                                       delimiter="~")
        task["milestone"] = self._get_last_chunk_of_url(raw["milestone_link"])
        bug_target_name = raw["bug_target_name"]

        if bug_target_name:
            task["target_name"] = bug_target_name.split("/")[0]
        else:
            task["target_name"] = None

        bug = self.get_bug_by_url(raw["bug_link"])
        to_copy = select_keys(bug, ["id", "title", "tags",
                                    "information_type", "private",
                                    "date_created", "date_last_updated"])
        task.update(to_copy)
        task["title"] = task["title"]

        task["owner"] = self._get_last_chunk_of_url(bug["owner_link"],
                                                    delimiter="~")

        task["verifier"] = self._process_bug_verifier(raw, bug)
        task["cves"] = self._get_bug_cves(task["id"])
        self._process_bug_dates(task)

        return task

    @cachetools.ttl_cache()
    def _get_json_from_url(self, url):
        resp = None
        try:
            resp = self.session.get(url, headers=self._create_auth_headers())
            return resp.json()
        except Exception as e:
            log.error("Failed to read from url {0}".format(url))
            log.error("Got response {0}".format(resp))

            log.error("Got bad response from LP "
                      "<{0}> - {1}".format(resp.status_code, resp.text))

            log.error(e, exc_info=True)
            raise e

    def _create_auth_headers(self):
        auth_header = collections.OrderedDict()
        auth_header["OAuth realm"] = "OAuth"
        auth_header["oauth_nonce"] = random.randint(10000000, 99999999)
        auth_header["oauth_timestamp"] = int(time.time())
        auth_header["oauth_consumer_key"] = self.oauth_consumer_key
        auth_header["oauth_signature_method"] = "PLAINTEXT"
        auth_header["oauth_version"] = "1.0"
        auth_header["oauth_token"] = self.oauth_token
        auth_header["oauth_signature"] = self.oauth_signature

        user_agent_header = collections.OrderedDict()
        user_agent_header["application"] = "lp-reports"
        user_agent_header["oauth_consumer"] = self.oauth_consumer_key

        headers = {
            "Authorization": self._create_header_string(auth_header),
            "user-agent": self._create_header_string(user_agent_header)
        }

        return headers

    @staticmethod
    def _create_header_string(options):
        string = ""
        for k, v in options.items():
            kv = "{0}=\"{1}\", ".format(k, v)
            string += kv

        # return everything but trailing comma and space
        return string[:-2]

    @staticmethod
    def _process_lp_date(bug_date):
        if bug_date:
            return iso8601.parse_date(bug_date)
        else:
            return None

    @staticmethod
    def _get_last_chunk_of_url(link, delimiter="/"):
        if link:
            return link.split(delimiter)[-1]
        else:
            return None
