import abc
import datetime
import itertools

import six

import pytz


class Predicate(object):
    six.add_metaclass(abc.ABCMeta)

    def __init__(self):
        pass

    @abc.abstractmethod
    def apply(self, bug):
        pass


class MembershipPredicate(Predicate):
    def __init__(self, bug_attr, values):
        super(MembershipPredicate, self).__init__()
        self.bug_attr = bug_attr
        self.values = values

    def apply(self, bug):
        return getattr(bug, self.bug_attr) in self.values


class TagMembershipPredicate(Predicate):
    def __init__(self, values, inclusive=True):
        super(TagMembershipPredicate, self).__init__()
        self.values = values
        self.inclusive = inclusive

    def apply(self, bug):
        found_len = len(self.intersect_lists(bug.tags, self.values))

        if self.inclusive:
            return found_len > 0
        else:
            return found_len == 0

    @staticmethod
    def intersect_lists(xs, ys):
        return set(xs).intersection(set(ys))


class TeamMembershipPredicate(Predicate):
    def __init__(self, bug_attr, values, teams_data):
        super(TeamMembershipPredicate, self).__init__()
        if bug_attr == 'reporter':
            self.bug_attr = 'owner'
        else:
            self.bug_attr = bug_attr
        self.values = values
        self.all_people = self.flatten_teams(teams_data)
        self.teams = self.extract_real_teams(teams_data)

    def _is_unknown(self, value):
        return ('unknown' in self.values) and (value not in self.all_people)

    def _is_in_team(self, value, team, members):
        return team in self.values and (value == team or value in members)

    def apply(self, bug):
        bug_attr = getattr(bug, self.bug_attr)

        if self._is_unknown(bug_attr):
            return True

        for team, members in self.teams.items():
            if self._is_in_team(bug_attr, team, members):
                return True

        return False

    @staticmethod
    def flatten_teams(major_teams):
        flatten = set()

        for team in major_teams.values():
            for subteam, people in team.items():
                flatten.add(subteam)
                flatten = flatten.union(people)

        return flatten

    @staticmethod
    def extract_real_teams(major_teams):
        teams = {}
        for t in major_teams.values():
            teams.update(t)

        return teams

    @staticmethod
    def is_teams_filter(name):
        return name in ['reporter', 'assignee', 'verifier']


class DateRangePredicate(Predicate):
    def __init__(self, predicate_name, values):
        super(DateRangePredicate, self).__init__()
        self.attr = self.transform_predicate_name(predicate_name)
        self.postfix = predicate_name.split("_")[-1]
        self.value = self.transform_date(values)

    def apply(self, bug):
        date = getattr(bug, self.attr)

        if not date:
            return False

        date = date.replace(tzinfo=pytz.UTC)

        if self.postfix == 'to':
            return date < self.value
        elif self.postfix == 'from':
            return date > self.value
        else:
            return False

    @staticmethod
    def transform_predicate_name(predicate_name):
        return 'date_' + predicate_name. \
            replace('_to', ''). \
            replace('_from', '')

    @staticmethod
    def transform_date(date_str):
        if date_str is not None:
            return datetime.datetime.strptime(date_str, "%Y-%m-%d").replace(tzinfo=pytz.UTC)

    @staticmethod
    def is_date_filter(name):
        date_fields = ["created", "triaged", "fix_committed", "fix_released"]
        return name in ([s + "_from" for s in date_fields] +
                        [s + "_to" for s in date_fields])


def compile_predicates(params, teams_data):
    predicates = []
    for name, values in params.items():
        pred = None

        if DateRangePredicate.is_date_filter(name):
            pred = DateRangePredicate(name, values)
        elif TeamMembershipPredicate.is_teams_filter(name):
            pred = TeamMembershipPredicate(name, values, teams_data)
        elif name == 'tags':
            pred = TagMembershipPredicate(values, inclusive=True)
        elif name == 'exclude_tags':
            pred = TagMembershipPredicate(values, inclusive=False)
        elif name == 'assigned_person':
            pred = MembershipPredicate('assignee', values)
        elif values:
            pred = MembershipPredicate(name, values)

        if pred:
            predicates.append(pred.apply)

    return lambda x: all(fn(x) for fn in predicates)


def prepare_query(query):
    # get rid of all the params with empty predicates
    return dict([i for i in query.items() if i[1]])


def find_proper_team_for_bug(bug, teams_data):
    # return first matching team
    return next(
        (team for team, members in teams_data.iteritems() if
         bug.assignee in itertools.chain.from_iterable(members.itervalues())),
        'Unknown')


def filter_bugs(bugs, search_params, teams_data):
    params = prepare_query(search_params)

    predicates = compile_predicates(params, teams_data)

    bugs['DATA'][0]['bugs'] = filter(predicates, bugs['DATA'][0]['bugs'])

    # group by teams
    if len(bugs['DATA']) == 2 and bugs['DATA'][1]:
        teams = bugs['DATA'][1]

        for bug in bugs['DATA'][0]['bugs']:
            team = find_proper_team_for_bug(bug, teams_data)
            teams[team]['bugs'].append(bug)

        bugs['DATA'] = teams.values()

    return bugs
