import cStringIO
import csv
from datetime import datetime

import flask
from flask import current_app as app
from flask import g

from lpreports import auth
from lpreports import db
from lpreports.launchpad import bug as lp_bug
from lpreports.launchpad import sla_reports
from lpreports import util

reports = flask.Blueprint('reports', __name__)


def _make_csv(bugs, filters):
    header = lp_bug.FIELDS_TO_COPY[:]
    header.append("team")

    rows = [header]
    for team_data in bugs['DATA']:
        for bug in team_data['bugs']:
            # There are non-ascii characters in bug titles
            # and we need to handle that explicitly
            title = bug.title.encode('ascii', 'ignore')

            row = []
            for field in lp_bug.FIELDS_TO_COPY:
                if field == "tags":
                    tags = getattr(bug, "tags") or []
                    row.append(" ".join(tags))
                elif field == "title":
                    row.append(title)
                else:
                    row.append(getattr(bug, field))

            row.append(team_data["display_name"])
            rows.append(row)

    str_io = cStringIO.StringIO()
    csvwriter = csv.writer(str_io)
    csvwriter.writerows(rows)

    return flask.Response(str_io.getvalue(),
                          mimetype='text/plain',
                          headers={"Content-Disposition":
                                   "attachment;filename=report.csv"})


def _render_custom_report(report):
    milestone_names = flask.request.args.getlist('milestone')

    projects = flask.request.args.getlist('project')
    if not projects:
        projects = ['mos', 'fuel']

    user = g.user
    security_membership = db.get_user_security_teams_membership(user["email"])

    bugs = sla_reports.get_reports_data(report,
                                        projects,
                                        milestone_names,
                                        security_membership)

    teams_data = db.get_full_teams()

    bugs, filters = util.filter_bugs(flask.request, bugs, teams_data)

    actual_milestones = db.get_milestones()
    if report['name'] == 'maintenance_report':
        actual_milestones = db.get_maintenance_milestones()
        bugs = process_maintenance_bugs(bugs)

    if flask.request.args.get('format', 'html') == 'csv':
        return _make_csv(bugs, filters)

    return flask.render_template(
        "report.html",
        report=report,
        all_bugs=bugs,
        teams=db.get_teams(),
        filters=filters,
        milestone_names=milestone_names,
        milestones=actual_milestones
    )


def process_maintenance_bugs(bugs):
    nearest_milestone = get_nearest_targeted_milestone()

    for bug in bugs['DATA'][0]['bugs']:
        assign_weight(bug, nearest_milestone)

    bugs_to_sort = bugs['DATA'][0]['bugs']
    bugs['DATA'][0]['bugs'] = sorted(bugs_to_sort,
                                     key=lambda bug: bug.weight,
                                     reverse=True)

    return bugs


def get_nearest_targeted_milestone():
    today = datetime.today()
    nearest_milestone = None
    target_dates = db.get_milestones_target_dates().items()
    if target_dates:
        nearest_milestone, nearest_target_date = target_dates[0]
        nearest_target_date = nearest_target_date.replace(tzinfo=None)
        least_delta = nearest_target_date - today
        for milestone, target_date in target_dates[1:]:
            target_date = target_date.replace(tzinfo=None)
            delta = target_date - today
            if delta < least_delta and delta.days >= 0:
                least_delta = delta
                nearest_milestone = milestone

    return nearest_milestone


def assign_weight(bug, nearest_milestone):
    weight = 0
    legend = []
    is_security_bug = 'Security' in bug.information_type
    if is_security_bug and bug.importance == "Critical":
        weight += 100
        legend.append((100, "Critical + Security"))
    if bug.importance == "Critical" and not is_security_bug:
        weight += 40
        legend.append((40, "Critical"))
    if bug.importance == "High":
        weight += 30
        legend.append((30, "High"))
    if bug.importance == "Medium":
        weight += 20
        legend.append((20, "Medium"))
    if bug.importance == "Low":
        weight += 10
        legend.append((10, "Low"))
    if 'customer-found' in bug.tags:
        weight += 10
        legend.append((10, "Customer found"))
    if is_security_bug and bug.importance != "Critical":
        weight += 5
        legend.append((5, "Security"))
    if 'sla1' in bug.tags:
        weight += 200
        legend.append((200, "SLA1"))
    if 'sla2' in bug.tags:
        weight += 10
        legend.append((10, "SLA2"))
    if nearest_milestone and bug.milestone == nearest_milestone:
        weight += 100
        legend.append((100, "within nearest milestone"))
    bug.set_weight(weight)
    bug.set_legend(legend)


@reports.route("/<report_name>")
@auth.requires_auth
def show(report_name):
    report = util.get_report_by_name(report_name)
    if not report:
        return flask.abort(404)

    return _render_custom_report(report)
